import hashlib
import json
import os
import ssl
import urllib

import urllib2

import time
import binascii
from Crypto.Cipher import AES

AGENT_SECRET = b'q6pdqy84mzs6ryf1'
AGENT_SCB = b'AQF90ZMqZAwc7_qR'
AGENT_IDENTIFIER = b'wmbns08zb9xucbmp'

DEBUG = True


class PassmanException(Exception):
    pass


class Passman:
    def __init__(self, config):
        self._access_token = None
        self._agent_auth = 'null'

        self._passman_url = config.get('passman', 'url')
        self._passman_ca_cert = config.get('passman', 'ca')
        self._passman_check_host_name = config.getboolean('passman', 'check_host_name')

    def set_access_token(self, access_token):
        self._access_token = access_token

    def set_auth_token(self, auth_token):
        self._agent_auth = auth_token

    def get_account(self, target_username, target_host, under_group=None):
        target_selector = {'__eqci_name': target_host, '_include_subfolders': True}
        if under_group is not None:
            target_selector['_under_group'] = under_group

        target_ids = self._get('targets', target_selector)
        if len(target_ids) == 0:
            raise PassmanException("No such target_host '{}' under group '{}'".format(target_host, under_group))

        if DEBUG:
            print('target ids', target_ids)

        # maybe get whole account to check if credential type is available
        accounts = self._get('accounts',
                             {'target': target_ids[0], 'name': target_username,
                              '_expand': True, '_expand_target': True, '_include_subfolders': True})
        if len(accounts['accounts']) != 1:
            raise PassmanException("No such target_username '{}'".format(target_username))

        return accounts['accounts'][0]

    def checkout_credential(self, account, credential_type):
        job_data = {
            "account": account['id'],
            "credential_type": credential_type
        }

        is_exclusive = False
        if account['target']['permitted_checkout_types'] == 'exclusive only':
            job_data['exclusive_duration'] = account['target']['max_exclusive_duration']
            is_exclusive = True

        checkout_job = self._post('jobs/LoginCheckoutCredentials', job_data)
        if checkout_job is not None and 'result' in checkout_job and credential_type in checkout_job['result']:
            return checkout_job['result'][credential_type], is_exclusive

        raise PassmanException('Error checking out credential {}'.format(checkout_job))

    def checkin_credentials(self, account_id):
        job_data = {"account": account_id}
        self._post('jobs/CheckinCredentials', job_data)

    def authenticate(self, username, password, provider, domain):
        auth_data = {
            'username': username,
            'password': password,
            'grant_type': 'password',
            'provider': provider
        }
        req = urllib2.Request(self._build_url('authorize'), data=urllib.urlencode(auth_data))
        response = self._do_request(req)
        if response is None:
            raise PassmanException("Authentication failed")

        return response['access_token'], self._create_auth_token(username, provider, domain)

    def whoami(self):
        return self._get('whoami')

    def _create_auth_token(self, username, provider, domain):
        message = 'un:' + username + ' iss:' + provider + ' dom:' + domain + ' ua:scb rt:' + str(time.time()) + ' '
        message += (16 - len(message) % 16) * '/'
        iv = os.urandom(16)
        cipher = AES.new(AGENT_SECRET, AES.MODE_CBC, iv)
        token_info = cipher.encrypt(message.encode())
        result = AGENT_IDENTIFIER + hashlib.sha256(AGENT_SCB).hexdigest().encode() + binascii.hexlify(token_info) + \
            binascii.hexlify(iv)
        return result.decode()

    def _get(self, resource, parameters=None):
        headers = self._build_authorization_header()
        req = urllib2.Request(self._build_url(resource, parameters), headers=headers)
        return self._do_request(req)

    def _post(self, resource, post_data, parameters=None):
        headers = self._build_authorization_header()
        headers['content-type'] = 'application/json'
        req = urllib2.Request(self._build_url(resource, parameters), data=json.dumps(post_data), headers=headers)
        return self._do_request(req)

    def _do_request(self, req):
        try:
            response = urllib2.urlopen(req, context=self._get_ssl_context())
            response_body = response.read()
            return json.loads(response_body)
        except urllib2.HTTPError as exc:
            print("Request failed", exc, exc.fp.read())
            return None

    def _get_ssl_context(self):
        ssl_context = ssl.create_default_context(cadata=self._passman_ca_cert)
        if self._passman_ca_cert is None:
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE
        else:
            ssl_context.verify_mode = ssl.CERT_REQUIRED
            ssl_context.check_hostname = self._passman_check_host_name

        return ssl_context

    def _build_authorization_header(self):
        if self._access_token is None:
            raise PassmanException("Missing access token, needs to authenticate first")

        headers = {
            'authorization': "Bearer {}".format(self._access_token), 'Agent-auth': self._agent_auth
        }
        return headers

    def _build_url(self, resource, parameters=None):
        if parameters is not None:
            params = "/?" + "&".join("{}={}".format(k, v) for k, v in parameters.iteritems())
        else:
            params = ""

        return "{}/api/{}{}".format(self._passman_url, resource, params)
